# import warnings
# warnings.filterwarnings("ignore", "Wswiglal-redir-stdio")

import h5py
import lal
import lalsimulation as lalsim
import numpy as np
import numpy.typing as npt
import phenom
from scipy.interpolate import InterpolatedUnivariateSpline as IUS

try:
    import pyseobnr  # pyright: ignore

    PYSEOBNR_AVAILABLE = True
except ImportError:
    PYSEOBNR_AVAILABLE = False


def td_amp_scale(mtot: float | int, distance: float | int) -> float:
    """
    Computes the amplitude pre-factor for time-domain signals
    given as M*G/c^2 * M_sun / dist

    Parameters
    ----------
    mtot : float | int
        Total mass in solar masses
    distance : float | int
        Distance to source in SI units (metres)

    Returns
    -------
    float
        The scale factor
    """
    return mtot * lal.MRSUN_SI / distance  # pyright: ignore


def gen_td_modes_wf_params(
    m1: int | float = 50,
    m2: int | float = 50,
    S1x: int | float = 0,
    S1y: int | float = 0,
    S1z: int | float = 0,
    S2x: int | float = 0,
    S2y: int | float = 0,
    S2z: int | float = 0,
    distance: int | float = 1,
    deltaT: int | float = 1 / 4096,
    phiRef: int | float = 0.0,
    f_min: int | float = 10,
    f_ref: int | float = 10,
    LALpars: None | lal.Dict = None,
    approximant: str = "SEOBNRv4P",
    lmax_dummy: int = 4,
) -> dict:
    """
    Function to convert the input into a dictionary that can be read by the waveform generator. The
    waveform generator in this case is either `gen_td_modes_wf` which calls
    `SimInspiralChooseTDModes` or `generate_td_modes` from `pyseobnr`.

    Parameters
    ----------
    m1: int | float, default 50
    m2: int | float, default 50
    S1x: int | float, default 0
    S1y: int | float, default 0
    S1z: int | float, default 0
    S2x: int | float, default 0
    S2y: int | float, default 0
    S2z: int | float, default 0
    distance: int | float, default 1
        distance in metres to the source
    deltaT: int | float, default 1 / 4096
    phiRef: int | float, default 0.0
    f_min: int | float, default 10
    f_ref: int | float, default 10
    LALpars: None | lal.Dict, default None
        Used to supply additional parameters to the generator.
        Typically used to set which modes to generate.
    approximant : str, default "SEOBNRv4P"
        The lalsim name of the approximant to generate or if generating
        using pyseobnr then this is either "SEOBNRv5HM" or "SEOBNRv5PHM"
    lmax_dummy : int, default 4
        old option for XLALSimInspiralChooseTDModes
        that sets the max ell mode, this tends to not be used
        as we use ModeArray instead in LALpars.

    Returns
    -------
    dict
        A dictionary with keys that the waveform generator expects.
    """

    if approximant in ["SEOBNRv5HM", "SEOBNRv5PHM"]:
        assert PYSEOBNR_AVAILABLE is True, "pyseobnr not installed"
        # pyseobnr
        distance_Mpc = distance / lal.PC_SI / 1e6
        p = {
            "mass1": m1,
            "mass2": m2,
            "spin1x": S1x,
            "spin1y": S1y,
            "spin1z": S1z,
            "spin2x": S2x,
            "spin2y": S2y,
            "spin2z": S2z,
            "deltaT": deltaT,
            "f22_start": f_min,
            "phi_ref": phiRef,
            "distance": distance_Mpc,
            "inclination": 0,
            # "f_max": 0.5 / deltaT, # this is the default value in pyseobnr i.e. Nyquist
            "approximant": approximant,
        }

    else:
        # lalsuite
        p = dict(
            m1=m1,
            m2=m2,
            S1x=S1x,
            S1y=S1y,
            S1z=S1z,
            S2x=S2x,
            S2y=S2y,
            S2z=S2z,
            phiRef=phiRef,
            r=distance,  # in metres
            deltaT=deltaT,
            f_min=f_min,
            f_ref=f_ref,
            LALpars=LALpars,
            lmax=lmax_dummy,
            approximant=lalsim.GetApproximantFromString(approximant),
        )

    return p


def gen_td_modes_wf(p: dict, modes: list[tuple[int, int]] = [(2, 2)]) -> tuple[npt.NDArray, dict]:
    """
    Function to generate waveform modes using lalsimulation's
    SimInspiralChooseTDModes function.

    Parameters
    ----------
        p : dict
            normally the output of gen_td_modes_wf_params
        modes : list[tuple[int,int]], default [(2,2)]
            modes to generate
            Note: Depending on the waveform model used
            you might need to explicitly provide both positive and
            negative modes.
            To supply multiple modes you do [(2,2), (2,1), (3,3), ...]

    Returns
    -------
        times : npt.NDArray
            The time grid that the waveform is evaluated at in units of seconds
        hlms: dict
            contains time domain complex hlm modes. The keys are given by the
            input modes.
    """
    p = p.copy()

    if (2, 2) not in modes:
        raise NotImplementedError(
            "(2,2) mode not in modes.\
Currently we assume that this mode exists."
        )

    if p["LALpars"] is None:
        p["LALpars"] = lal.CreateDict()

    # amplitude order?
    # lalsim.SimInspiralWaveformParamsInsertPNAmplitudeOrder(p['LALpars'], -1)

    ma = lalsim.SimInspiralCreateModeArray()
    for ell, mm in modes:
        lalsim.SimInspiralModeArrayActivateMode(ma, ell, mm)
    lalsim.SimInspiralWaveformParamsInsertModeArray(p["LALpars"], ma)

    # M = p["m1"] + p["m2"]
    p.update({"m1": p["m1"] * lal.MSUN_SI})
    p.update({"m2": p["m2"] * lal.MSUN_SI})

    hlms_lal = lalsim.SimInspiralChooseTDModes(**p)

    hlms = {}

    for ell, mm in modes:
        tmp = lalsim.SphHarmTimeSeriesGetMode(hlms_lal, ell, mm)
        if ell == 2 and mm == 2:
            length_22 = tmp.data.length
            dt_22 = tmp.deltaT
            epoch_22 = tmp.epoch
        hlms.update({(ell, mm): tmp.data.data})

    assert (
        p["deltaT"] == dt_22  # pyright: ignore
    ), f"input deltaT = {p['deltaT']} does not match waveform dt = {dt_22}."  # pyright: ignore

    t = np.arange(length_22) * dt_22 + float(epoch_22)  # pyright: ignore

    return t, hlms


def get_hdf5_strain(nr_hdf5_filename: str, modes: list[tuple[int, int]], dt: int | float) -> dict:
    """
    Load NR modes from a LAL compatible hdf5 file and return the output as a dictionary.
    The data is interpolated onto a time grid with spacing given by the input dt in units
    of M (total mass).

    Parameters
    ----------
    nr_hdf5_filename: str
        The path the the NR hdf5 file to load.
    modes: list[tuple[int,int]]
        List of modes to load. e.g. [(2,2), (2,-2), (2, 1), (2, -1), ...]
    dt: int | float
        The grid spacing for use for the output time array in units of M (total mass)

    Returns
    -------
    dict
        The output dict has the following keys:
            - t: npt.NDArray
                The times the waveform is defined at in units of M (total mass)
            - hlm:dict
                Contains time domain complex hlm modes. The keys are given by the
                input modes.
            - metadata:dict
                A dictionary that contains some of the NR metadata such as masses,
                spins and filename.
    """
    f = h5py.File(nr_hdf5_filename, "r")

    # eta = f.attrs["eta"]
    # this try/except is here because the GTech waveforms
    # use 'irreducible_mass1' instead of 'mass1'..
    try:
        q = float(f.attrs["mass1"]) / float(f.attrs["mass2"])
        Mtotal = float(f.attrs["mass1"]) + float(f.attrs["mass2"])
    except KeyError:
        q = float(f.attrs["irreducible_mass1"]) / float(f.attrs["irreducible_mass2"])
        Mtotal = float(f.attrs["irreducible_mass1"]) + float(f.attrs["irreducible_mass2"])

    # print((f.attrs.keys()))
    spin1z = f.attrs["spin1z"]
    spin2z = f.attrs["spin2z"]

    amp_x = {}
    amp_y = {}
    phase_x = {}
    phase_y = {}
    for lm in modes:
        amp_tmp = f["amp_l{0}_m{1}".format(lm[0], lm[1])]
        amp_x[lm] = amp_tmp["X"][()]
        amp_y[lm] = amp_tmp["Y"][()]

        phase_tmp = f["phase_l{0}_m{1}".format(lm[0], lm[1])]
        phase_x[lm] = phase_tmp["X"][()]
        phase_y[lm] = phase_tmp["Y"][()]

    f.close()

    t1 = max(amp_x[2, 2][0], phase_x[2, 2][0])
    t2 = min(amp_x[2, 2][-1], phase_x[2, 2][-1])

    times = np.arange(t1, t2, dt)

    hlm = {}
    for lm in modes:
        amp_i = IUS(amp_x[lm], amp_y[lm])
        phase_i = IUS(phase_x[lm], phase_y[lm])

        amp = amp_i(times)
        phase = phase_i(times)

        hlm[lm] = amp * np.exp(1.0j * phase)

    # TODO: output f.attrs as metadata
    metadata = {
        "q": q,
        "filename": nr_hdf5_filename,
        "M": Mtotal,
        "spin1z": spin1z,
        "spin2z": spin2z,
    }
    wf = dict(t=times, hlm=hlm, metadata=metadata)
    return wf


def generate_waveform(
    q: int | float,
    modes: list[tuple[int, int]],
    M: int | float = 50,
    f_min: int | float = 30,
    S1z: int | float = 0,
    S2z: int | float = 0,
    approximant: str = "SpinTaylorT1",
    deltaT: int | float = 1 / 4096,
    phiRef: int | float = 0,
    f_ref: int | float | None = None,
):
    """
    Main function to generate waveforms.
    Returns a dictionary with keys 't' and 'hlm' where
    'hlm' is a dictionary with keys (l,m).
    the time is in units of M. The hlm data are also normalised
    by the time domain scaling factor.

    Parameters
    ----------
    q: int | float
        The mass-ratio. We use the convention that q >= 1.
    modes: list[tuple[int, int]],
        list of [(l,m)] modes to include. E.g. [(2,2), (2,1), (3,3), ...]
    M: int | float, default 50
        total mass in solar masses
    f_min: int | float, default 30
        start frequency in Hz
    S1z: int | float, default 0
        Z component of body 1 spin
    S2z: int | float, default 0
        Z component of body 2 spin
    approximant: str, default "SpinTaylorT1"
        The name of the approximant. Usually the lalsimulation name.
        If either "SEOBNRv5HM" or "SEOBNRv5PHM" provided then this generated using
        pyseobnr.
    deltaT: int | float, default 1 / 4096
        The time spacing to generate the waveform at. Also the inverse sample rate.
    phiRef: int | float, default 0
        The reference phase. This can be defined differently depending on the approximant.
        But it typically should be the value of the 2,2 GW phase at the reference frequency.
    f_ref: int | float | None, default None
        The reference frequency in Hz. If None then this is set to f_min.


    Returns
    -------
    dict
        The output dict has the following keys:
            - t: npt.NDArray
                The times the waveform is defined at in units of M (total mass)
            - hlm : dict
                Contains time domain complex hlm modes. The keys are given by the
                input modes. The amplitude of the hlm data have been divided by the
                comman time domain amplitude scaling factor.
    """
    assert q >= 1, f"Input mass-ratio q is {q}. We use the convention that q >= 1"
    if f_ref is None:
        f_ref = f_min
    m1, m2 = phenom.m1_m2_M_q(M, q)
    p = gen_td_modes_wf_params(
        approximant=approximant,
        m1=m1,
        m2=m2,
        f_min=f_min,
        f_ref=f_ref,
        phiRef=phiRef,
        S1z=S1z,
        S2z=S2z,
        deltaT=deltaT,
    )

    if approximant in ["SEOBNRv5HM", "SEOBNRv5PHM"]:
        assert PYSEOBNR_AVAILABLE is True, "pyseobnr not installed"
        # pyseobnr
        p.update({"mode_array": modes})
        wfm_gen = pyseobnr.generate_waveform.GenerateWaveform(p)
        # Generate mode dictionary
        t, hlm = wfm_gen.generate_td_modes()
        # convert Mpc to metres
        r = p["distance"] * lal.PC_SI * 1e6
    else:
        # lalsuite
        t, hlm = gen_td_modes_wf(p, modes=modes)
        r = p["r"]

    amp_scale = td_amp_scale(M, r)  # pyright: ignore
    for lm in hlm.keys():
        hlm[lm] = hlm[lm] / amp_scale
    t_M = phenom.StoM(t, M)
    wf = dict(t=t_M, hlm=hlm)
    return wf



"""
This module implements the aligned-spin TaylorT3 equations.

Some references:

 - [arXiv:0802.1249](https://arxiv.org/abs/0802.1249)
 - [arXiv:0901.2437](https://arxiv.org/abs/0901.2437)
 - [arXiv:0610122](https://arxiv.org/abs/gr-qc/0610122)
 - [arXiv:0907.0700](https://arxiv.org/abs/0907.0700)
 - [arXiv:0406012](https://arxiv.org/abs/gr-qc/0406012)
 - [arXiv:1210.2339](https://arxiv.org/abs/1210.2339)
 - [arXiv:1210.2339](https://arxiv.org/abs/1210.2339)
 - [arXiv:2004.08302](http://arxiv.org/abs/2004.08302)
 - [arXiv:2012.11923](http://arxiv.org/abs/2012.11923)
 - [LALSimInspiralTaylorT3.c](https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/lib/LALSimInspiralTaylorT3.c)
 - [LALSimInspiralPNCoefficients.c](https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c)
"""

import numpy as np
from lal import GAMMA, MTSUN_SI


def Msun_to_sec(M: float | int) -> float:
    """
    Geometric units convertion.
    convert mass (in units of solar masses)
    into seconds.

    The equation is M * MSUN_SI * G_SI / C_SI**3.

    Parameters
    ----------
    M : float | int
        Total mass in solar masses

    Returns
    -------
    float | int
        Mass converted into units of seconds.
    """
    #     return
    return M * MTSUN_SI


def TaylorT3_Omega_GW_Newt(t, tc: float | int, eta: float, M: float | int):
    """
    Newtonian pre factor for GW angular frequency

    Parameters
    ----------
    t : number or array
        The time coordinate
    tc : float | int
        The TaylorT3 coalescence time
    eta : float
        The symmetric mass ratio
    M : float | int
        Total mass in solar masses

    Returns
    -------
    float or array
        Newtonian pre factor for GW angular frequency
    """
    theta = TaylorT3_theta(t, tc, eta, M)
    return theta**3 / (8.0 * M)


def TaylorT3_theta(t, tc, eta, M):
    """
    Taylor3 parameter
    """
    theta = eta * (tc - t) / (5 * M)
    theta = theta ** (-1 / 8.0)
    return theta


def TaylorT3_t(theta, tc, eta, M):
    """
    Inverse of TaylorT3_theta
    """
    return -5 * M / eta / theta**8 + tc


def TaylorT3_Omega_GW(t, tc, eta, M, chi1, chi2):
    """
    22 mode angular GW frequency
    equation 7 in 0901.2437

    3.5PN term from https://arxiv.org/pdf/gr-qc/0610122.pdf and https://arxiv.org/pdf/0907.0700.pdf
    and this too apparently https://arxiv.org/pdf/gr-qc/0406012.pdf?

    https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/lib/LALSimInspiralTaylorT3.c

    https://git.ligo.org/lscsoft/lalsuite/blob/master/lalsimulation/lib/LALSimInspiralPNCoefficients.c

    spinning terms from: http://arxiv.org/abs/2004.08302

    Parameters
    ----------
    t:
        time
    tc:
        coalescence time
    eta:
        symmetric mass ratio
    M:
        total mass (Msun)
    chi1:
        dimensionless spin of the primary (larger) black hole
    chi2:
        dimensionless spin of the secondary (smaller) black hole

    Returns
    -------
    """

    delta_m = np.sqrt(1 - 4 * eta)
    m1 = (1 + delta_m) / 2
    m2 = (1 - delta_m) / 2
    total_chi = chi1 + chi2
    m_i_chi_i = m1 * chi1 + m2 * chi2
    m_i_chi_i_sq = m1 * chi1**2 + m2 * chi2**2
    m_i_chi_i_cube = m1 * chi1**3 + m2 * chi2**3

    pi2 = np.pi * np.pi

    theta = TaylorT3_theta(t, tc, eta, M)

    theta2 = theta * theta
    theta3 = theta2 * theta
    theta4 = theta3 * theta
    theta5 = theta4 * theta
    theta6 = theta5 * theta
    theta7 = theta6 * theta

    # pre factor
    ftaN = 1.0 / (8.0 * M)
    # 0PN
    fts1 = 1.0
    # 0.5PN = 0 in GR
    # 1PN
    fta2 = 7.43 / 26.88 + 1.1 / 3.2 * eta
    # 1.5PN
    fta3_nonspin = -3.0 / 10.0 * np.pi
    fta3_spin = 113.0 / 160.0 * m_i_chi_i - 19.0 / 80.0 * eta * total_chi
    fta3 = fta3_nonspin + fta3_spin
    # 2PN
    fta4_nonspin = 1.855099 / 14.450688 + 5.6975 / 25.8048 * eta + 3.71 / 20.48 * eta * eta
    fta4_spin = -243.0 / 1024.0 * m_i_chi_i_sq + 3.0 * eta / 1024.0 * (
        81 * chi1**2 - 158 * chi1 * chi2 + 81 * chi2**2
    )
    fta4 = fta4_nonspin + fta4_spin
    # 2.5PN
    fta5_nonspin = (-7.729 / 21.504 + 1.3 / 25.6 * eta) * np.pi
    fta5_spin = (
        146597.0 / 64512.0 * m_i_chi_i
        - 1213.0 / 1152.0 * eta * total_chi
        + 7.0 / 128.0 * eta * delta_m * (chi1 - chi2)
        - 17.0 / 128.0 * eta * eta * total_chi
    )
    fta5 = fta5_nonspin + fta5_spin
    # 3PN
    fta6_nonspin = (
        -7.20817631400877 / 2.88412611379200
        + 5.3 / 20.0 * pi2
        + 1.07 / 2.80 * GAMMA
        + (25.302017977 / 4.161798144 - 4.51 / 20.48 * pi2) * eta
        - 3.0913 / 183.5008 * eta * eta
        + 2.35925 / 17.69472 * eta * eta * eta
    )
    fta6_spin = (
        -6127.0 / 6400.0 * np.pi * m_i_chi_i
        - 16928263.0 / 68812800.0 * m_i_chi_i_sq
        + 1051.0 / 3200.0 * np.pi * eta * total_chi
        + 23281001.0 / 68812800.0 * eta * (chi1**2 + chi2**2)
        - 377345.0 / 1376256.0 * eta * chi1 * chi2
        + 453767.0 / 4915200.0 * eta * delta_m * (chi1**2 - chi2**2)
        + 335129.0 / 2457600.0 * eta**2 * (chi1**2 + chi2**2)
        - 488071.0 / 1228800.0 * eta**2 * chi1 * chi2
    )
    fta6 = fta6_nonspin + fta6_spin

    # 3.5PN
    fta7_nonspin = (
        -1.88516689 / 4.33520640 - 9.7765 / 25.8048 * eta + 1.41769 / 12.90240 * eta * eta
    ) * np.pi
    fta7_spin = (
        6579635551.0 / 650280960.0 * m_i_chi_i
        + 3663.0 / 5120.0 * m_i_chi_i_sq
        - 67493.0 / 81920.0 * m_i_chi_i_cube
        - 1496368361.0 / 185794560.0 * eta * total_chi
        - 3663.0 / 5120.0 * np.pi * eta * (chi1**2 + chi2**2)
        + 3537.0 / 2560.0 * np.pi * eta * chi1 * chi2
        + 206917.0 / 163840.0 * eta * (chi1**3 + chi2**3)
        - 192709.0 / 81920.0 * eta * chi1 * chi2 * m_i_chi_i
        - 28633921.0 / 12386304.0 * eta * delta_m * (chi1 - chi2)
        + 71931.0 / 163840.0 * eta * delta_m * (chi1**3 - chi2**3)
        - 840149.0 / 15482880.0 * eta * eta * total_chi
        - 2219.0 / 40960.0 * eta * eta * (chi1**3 + chi2**3)
        + 1343.0 / 40960.0 * eta * eta * chi1 * chi2 * total_chi
        + 2591.0 / 46080.0 * eta * eta * delta_m * (chi1 - chi2)
        - 12029.0 / 92160.0 * eta * eta * eta * total_chi
    )
    fta7 = fta7_nonspin + fta7_spin

    # 3PN log term
    ftal6 = 1.07 / 2.80

    omega_orb = (
        theta3
        * ftaN
        * (
            fts1
            + fta2 * theta2
            + fta3 * theta3
            + fta4 * theta4
            + fta5 * theta5
            + (fta6 + ftal6 * np.log(2.0 * theta)) * theta6
            + fta7 * theta7
        )
    )

    # convert from orb to 22 GW
    return 2 * omega_orb


def Hhat22_pre_factor(x, eta):
    """
    https://arxiv.org/pdf/0802.1249.pdf
    eq. 9.3a and 9.3b
    """
    return np.sqrt(16.0 * np.pi / 5) * 2 * eta * x


def Hhat22_x(x, eta, chi1, chi2):
    """
    https://arxiv.org/pdf/0802.1249.pdf - eq. 9.4a

    3.5PN term: https://arxiv.org/pdf/1210.2339.pdf

    Spinning part from http://arxiv.org/abs/2004.08302 and http://arxiv.org/abs/2012.11923
    Here I implement the spinning terms from the 2nd paper (Equation A1) becuase I think
    the spinning terms from the 1st paper might have typos.
    Note that m1 = (1 + delta)/2 and m2 = (1 - delta)/2.

    here we leave the expression to depend on the post-newtonian
    parameter 'x' so that you can choose how to calculate it.
    e.g., from PN like TaylorT3 or from the model which
    is TaylorT3 + corrections

    return complex
    """
    delta_m = np.sqrt(1 - 4 * eta)

    xarr = np.zeros(8, dtype=np.complex128)

    xarr[0] = 1.0
    # 0.5 PN term is zero
    xarr[1] = 0
    xarr[2] = -107.0 / 42 + 55 * eta / 42
    x3_non_spin = 2.0 * np.pi

    x3_spin = 2.0 / 3.0 * (-delta_m * (chi1 - chi2) + eta * (chi1 + chi2) - (chi1 + chi2))

    xarr[3] = x3_non_spin + x3_spin
    x4_non_spin = -2173.0 / 1512 - 1069.0 * eta / 216 + 2047.0 * eta**2 / 1512

    chi1_2 = chi1**2
    chi2_2 = chi2**2
    x4_spin = (
        delta_m / 2 * (chi1_2 - chi2_2)
        - eta * (chi1_2 + chi2_2 - 2 * eta * chi1 * chi2)
        + (chi1_2 + chi2_2) / 2.0
    )

    xarr[4] = x4_non_spin + x4_spin
    xarr[5] = -107 * np.pi / 21 - 24.0 * 1.0j * eta + 34.0 * np.pi * eta / 21

    x6a = 27027409.0 / 646800 - 856.0 * GAMMA / 105 + 428 * 1.0j * np.pi / 105 + 2.0 * np.pi**2 / 3
    x6b = (
        (-278185.0 / 33264 + 41 * np.pi**2 / 96) * eta
        - 20261.0 * eta**2 / 2772
        + 114635.0 * eta**3 / 99792
    )

    x6log = -428.0 * np.log(16 * x) / 105

    xarr[6] = x6a + x6b

    xarr[7] = (
        -2173 * np.pi / 756
        + (-(2495 * np.pi / 378) + (14333 * 1.0j / 162)) * eta
        + ((40 * np.pi / 27) - (4066 * 1.0j / 945)) * eta**2
    )

    # pn = xarr[0] + x*xarr[2] + x**(3/2.)*xarr[3] + x**2*xarr[4] + x**(5/2.)*xarr[5] + x**3*(xarr[6] + x6log) + x**(7/2.)*xarr[7]
    pn = xarr[0]
    pn += x * xarr[2]
    pn += x ** (3 / 2.0) * xarr[3]
    pn += x**2 * xarr[4]
    pn += x ** (5 / 2.0) * xarr[5]
    pn += x**3 * (xarr[6] + x6log)
    pn += x ** (7 / 2.0) * xarr[7]

    pre = Hhat22_pre_factor(x, eta)

    return pre * pn


def x_from_omega_22(GW22AngFreq, M=1):
    OrgAngFreq = GW22AngFreq / 2
    x = (M * OrgAngFreq) ** (2.0 / 3)
    return x


def TaylorT3_Hhat22(t, tc, eta, M, chi1, chi2):
    """
    https://arxiv.org/pdf/0802.1249.pdf - eq. 9.4a
    Post-Newtonian expression for (l,m)=(2,2) time domain
    amplitude assuming TaylorT3 frequency evolution
    """

    GW22AngFreq = TaylorT3_Omega_GW(t, tc, eta, M, chi1, chi2)
    x = x_from_omega_22(GW22AngFreq, M)
    return Hhat22_x(x, eta, chi1, chi2)



import numpy as np
import numpy.typing as npt
import sympy
from scipy.interpolate import InterpolatedUnivariateSpline as IUS
from sympy import Add

# the only sympy symbol we use is 'x_sym'
x_sym = sympy.symbols("x_sym")


def compute_values(
    collocation_points: dict,
    x: npt.NDArray,
    y: npt.NDArray,
) -> dict:
    """
     Given a dictionary of collocation points compute the collocation values. x and y are training
     data on a fine grid which are interpolated so that derivatives can be estimated.

     Parameters
     ----------
     collocation_points : dict
         Collocation points supplied as a dictionary with integer keys representing the order of
         derivative and the values are a list of numbers that are the values of the x-coordinate to
         evaluate the (x,y) data at e.g. {0:[], 1:[], ...}.
     x : npt.NDArray
         Independent variable
    y : npt.NDArray
         Dependent variable

     Returns
     -------
     cv : dict
         A dictionary of collocation values (cv). The dictionary has the same
         format as the input collocation_points dict.
    """
    cp = collocation_points.copy()
    iy = IUS(x, y)
    cv = {d: np.asarray([iy.derivative(d)(v) for v in cp[d]]) for d in cp.keys()}
    return cv


class CollocationModel:
    """
    CollocationModel is a class that can fit a linear model to data
    using the collocation point method. The collocation points are
    not restricted to being just the value of the function at the collocation
    points but you can also specify the value to be higher derivatives
    of the model at the collocation points.

    For more details on this implementation see Section 5 in:
    [PhysRevD.109.104045](https://journals.aps.org/prd/abstract/10.1103/PhysRevD.109.104045).
    """

    def __init__(
        self,
        collocation_points: dict,
        collocation_values: dict,
        ansatz: Add,
        sub_dict: dict | None = None,
    ):
        """
        collocation_points : dict
            dictionary with integer keys representing the order
            of derivative and the values are a list of numbers that are the
            values of the x-coordinate to evaluate the rhs at.
            {0:[], 1:[], ...}
            [[d0], [d1], [d2], ...]
            d0 == points to evaluate RHS at zeroth derivative
            d1 == points to evaluate RHS at 1st derivative
            d2 == points to evaluate RHS at 2nd derivative ...etc
            The sum total of elments of the flattened list is the
            number of equations and equivalently the number of
            coefficients to fix by comparing with data.
            Data is provided as the right hand side (RHS).
        collocation_values : dict
            the collocation values are also called the right hand side (rhs)
            provided in the same format as collocation_points
        ansatz:
            sympy stuff
            this must be a linear-in-coefficients ansatz so that
            it can be expressed in a matrix form.
            Ansatz is just another word for functional form / model.
            For example a linear model/ansatz could be y(x) = a*x + b
        sub_dict:
            this is a dictionary mapping sympy symbol name to a constant
            this is needed because after we (potentially) differentiate the ansatz
            term we need to evaluate the expression so we need a value for these constants
            Note: currently we only support additionaly constants that linearly multiply the
            variables.
            e.g. if the ansatz is 1 + x then you could use sub_dict for ansatz of the form
            1 + x * omega but not for something like 1 + x + omega.
        """
        self.collocation_points = collocation_points
        # self.collocation_values = collocation_values
        self.rhs = collocation_values
        self.ansatz = ansatz.copy()
        self.sub_dict = sub_dict

        self.setup()

    def setup(self):
        # flatten collocation points
        self.collocation_points_1d = np.concatenate(
            [v for _, v in self.collocation_points.items()]
        )
        # set degrees of freedom / number of free parameters expected
        self.dof = len(self.collocation_points_1d)
        # create some human readable names for the collocation points
        # for example if collocation_points = {0:[1,2,3], 1:[2,3]}
        # then tags = ["d0_c1","d0_c2","d0_c3", "d1_c1", "d1_c2"]
        self.tags = []
        for d in self.collocation_points.keys():
            for c in range(len(self.collocation_points[d])):
                self.tags.append(f"d{d}_c{c}")

        self.lambdify_ansatz(sympy_args=(x_sym))

        # flatten right hand side
        self.rhs_1d = np.concatenate([v for _, v in self.rhs.items()])
        assert len(self.collocation_points_1d) == len(
            self.rhs_1d
        ), "number of collocation points and rhs are not the same"

        msg = f"""ansatz doesn't contain correct number of degrees of freedom.
 dof = {self.dof}. Ansatz has {len(self.ansatz.args)}"""
        assert self.dof == len(self.ansatz.args), msg

        self.create_information_matrix()
        self.solve()

    def lambdify_ansatz(self, sympy_args):
        """
        Create a sympy lambda function for each of the terms in the ansatz.
        sympy_args: these are the sympy symbols that that lambda function will
            be a function of
        """
        # terms_fn: list of sympy.lambdify functions
        # each element of the list corresponds to each term in the ansatz
        ansatz = self.ansatz.copy()
        if self.sub_dict is not None:
            ansatz = ansatz.subs(self.sub_dict)
        # self.terms_fn = [sympy.lambdify(sympy_args, term, "numpy") for term in self.ansatz.args]
        self.terms_fn = [sympy.lambdify(sympy_args, term, "numpy") for term in ansatz.args]

    def create_information_matrix(self):
        """
        Evaluate the ansatz at the collocation_points and create the information matrix. It's
        elements are the values of the variables (also called indeterminates) of the Ansatz
        evaluated at each of the collocation points.

        The information matrix is calculated using two for loops. The first loop is over
        the derivative order and the second loop is over collation points.
        Each iteration of this double for loop creates a row in the information matrix.
        For example if collocation_points = {0:[0,1], 1:[0]}
        then the rows in information_matrix matrix will correspond to
        [
            [derivative=0, collocation_point=0],
            [derivative=0, collocation_point=1],
            [derivative=1, collocation_point=0]
        ]

        extra_sub_dict: this is a dictionary mapping sympy symbol name to a constant
            this is needed because after we (potentially) differentiate the ansatz
            term we need to evaluate the expression so we need a value for these constants
        """
        # compute information matrix

        # need to compute derivatives term by term
        # that way we can be sure that we arrive at the correct number of
        # elements in each row
        # and then we can correctly handle when a derivative is zero.
        row = []
        # loop over groups of collocation points which corresponds to different derivative orders
        # for d in range(len(self.collocation_points)):
        for d in self.collocation_points.keys():
            # loop over each collocation point for each derivative
            for c in self.collocation_points[d]:
                # take derivative of each term in the ansatz
                terms = [sympy.diff(term, x_sym, d) for term in self.ansatz.args]
                # evaluate terms at given collocation point
                sub_dict = {"x_sym": c}
                # evaluate terms at given constants (if any)
                if self.sub_dict is not None:
                    sub_dict.update(self.sub_dict)
                row.append([term.subs(sub_dict) for term in terms])
        information_matrix = np.array(row, dtype=np.float64)
        self.information_matrix = information_matrix

    def solve(self, information_matrix=None, rhs=None):
        """
        information_matrix : None, default is None
            This matrix contains the value of variables
            (indeterminates) of the model/ansatz at the collocation
            points.
            If not given then uses the current one.
        rhs : None, default is None
            Flattened rhs dictionary. If not given then uses the current one.
        """
        if information_matrix is None:
            information_matrix = self.information_matrix
        if rhs is None:
            rhs = self.rhs_1d
        self.coeffs = np.linalg.solve(information_matrix, rhs)

    def get_basis(self, X: npt.NDArray, dtype=object):
        """
        Turn this into a property so that it only gets calculated
        once/when it needs to.

        X : npt.NDArray
            TODO
        dtype : object
            used to be object because in sympy when we have a constant
            term such as x_sym**0 this gets simplified to just '1'.
            This doesn't get broadcast when you pass an array but
            seems to work with dtype=object
        """
        basis = np.array([t(X) for t in self.terms_fn], dtype=dtype)
        return basis

    def predict(self, X: npt.NDArray, dtype=np.float64) -> npt.NDArray:
        # turn basis into a property that only gets
        # recalculated when the `x`s change.
        basis = self.get_basis(X)

        return np.dot(self.coeffs, basis).astype(dtype)



import copy

import lmfit
import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline as IUS

from prim import waveform


def MultimodeHybridModel(x, dt, dphi, dpsi, waveform_nr, modes):
    """
    This function is designed to be used with lmfit.
    It can fit multiple modes by concatenating them.

    Use this to help
    https://lmfit.github.io/lmfit-py/faq.html#how-can-i-fit-multiple-data-sets

    input:
        x: time arrray to evaluate model on
        dt: time shift parameter to estimate
        dphi: orbital phase shift parameter to estimate
        dpsi: polarisation angle parameter to estimate
        waveform_nr: an instance of the waveform.Waveform class
            containing the NR data.
            Specifically, all that is needed are the times
            'waveform_nr.times' and
            the dictionary complex multipoles 'waveform_nr.hlms'
        modes: list of (l,m) modes to use in the fit

    returns:
        array where all hlm_nr_y multipole data has been
        concatenated, selecting only the input `x` times
        with the time, phase and polarisation shift applied.
    """
    # assert type(waveform_nr) == waveform.Waveform, 'must be of type waveform.Waveform'
    # waveform_nr = copy.deepcopy(waveform_nr)
    i_hlm_real = {}
    i_hlm_imag = {}
    for k in modes:
        i_hlm_real[k] = IUS(waveform_nr.times, waveform_nr.hlms[k].real, ext=3)
        i_hlm_imag[k] = IUS(waveform_nr.times, waveform_nr.hlms[k].imag, ext=3)

    zs = []
    for k in modes:
        m = k[1]
        z = i_hlm_real[k](x + dt) + 1.0j * i_hlm_imag[k](x + dt)
        z *= np.exp(1.0j * m * dphi)
        z *= np.exp(1.0j * dpsi)
        zs.append(z)

    return np.concatenate(zs)


def fit_hybrid(wf_pn, wf_nr, win1, win2, modes, time_shift_guess, n_tries=5, max_nfev=None):
    """
    wf_pn: an instance of the waveform.Waveform class
            containing the PN data.
        This waveform could also be called waveform_1, the fixed waveform or the 'left' waveform.
        During the fitting process we leave this waveform fixed / untransformed.
        It's called wf_pn but the waveform doesn't have to be from PN
    wf_nr: an instance of the waveform.Waveform class
            containing the NR data.
        This is the 'right' waveform. In this code we apply transformations to this
        waveform to align it with wf_pn.
        It's called wf_nr but the waveform doesn't have to be from NR.
    win1: start time of hybrid window, in pn times
    win2: end time of hybrid window, in pn times
    modes: list of tuples of which modes to use to fit
    """
    wf_pn = copy.deepcopy(wf_pn)
    wf_nr = copy.deepcopy(wf_nr)

    ##############################
    # prepare PN data
    wf_pn.mask(start_time=win1, end_time=win2)
    x_fit = wf_pn.times
    y_fit = np.concatenate([wf_pn.hlms[k] for k in modes])
    #
    ##############################

    # need to fit each waveform multiple times
    # with either dpsi fixed to either 0 or pi
    # and then with random dphis to avoid local minima
    # pick the model with smallest redchis

    # it's 0 or pi because we define it as exp(i*psi0)

    # psi=0
    results_psi_0 = []
    chisq_psi_0 = []

    # psi=pi
    results_psi_pi = []
    chisq_psi_pi = []

    model = lmfit.Model(
        MultimodeHybridModel,
        independent_vars="x",
        param_names=["dt", "dphi", "dpsi"],
        waveform_nr=wf_nr,
        modes=modes,
    )

    # parallelise this loop?
    for i in range(n_tries):
        # psi=0
        params = model.make_params(
            dt=dict(value=time_shift_guess),
            dphi=dict(value=np.random.uniform(0, 2 * np.pi), min=0, max=2 * np.pi),
            dpsi=dict(value=0, vary=False),
        )

        result = model.fit(y_fit, params, x=x_fit, max_nfev=max_nfev)
        results_psi_0.append(result)
        chisq_psi_0.append(result.redchi)

        # psi=pi
        params = model.make_params(
            dt=dict(value=time_shift_guess),
            dphi=dict(value=np.random.uniform(0, 2 * np.pi), min=0, max=2 * np.pi),
            dpsi=dict(value=np.pi, vary=False),
        )

        result = model.fit(y_fit, params, x=x_fit, max_nfev=max_nfev)
        results_psi_pi.append(result)
        chisq_psi_pi.append(result.redchi)

    psi_0_best_idx = np.argmin(chisq_psi_0)
    psi_0_best_value = chisq_psi_0[psi_0_best_idx]

    psi_pi_best_idx = np.argmin(chisq_psi_pi)
    psi_pi_best_value = chisq_psi_pi[psi_pi_best_idx]

    # return model with smallest chisquare
    if psi_0_best_value <= psi_pi_best_value:
        return results_psi_0[psi_0_best_idx]
    else:
        return results_psi_pi[psi_pi_best_idx]


def get_window_times(wf, n_cycles_before_window=2, n_cycles_in_window=6):
    wf = copy.deepcopy(wf)
    wf.compute_phase()

    # estimate derivative of phase using the first half of the data
    idx_half = int(len(wf.phases[2, 2]) // 2)
    phi = wf.phases[2, 2][:idx_half]
    avg_sign = np.mean(np.diff(phi))
    if avg_sign < 0:
        phi_win1 = wf.phases[2, 2][0] - 2 * np.pi * n_cycles_before_window
        phi_win2 = phi_win1 - 2 * np.pi * n_cycles_in_window
        win1 = wf.times[wf.phases[2, 2] >= phi_win1][-1]
        win2 = wf.times[wf.phases[2, 2] >= phi_win2][-1]
    else:
        phi_win1 = wf.phases[2, 2][0] + 2 * np.pi * n_cycles_before_window
        phi_win2 = phi_win1 + 2 * np.pi * n_cycles_in_window
        win1 = wf.times[wf.phases[2, 2] <= phi_win1][-1]
        win2 = wf.times[wf.phases[2, 2] <= phi_win2][-1]

    return win1, win2


def blending_function(t, win1, win2):
    """
    https://arxiv.org/abs/1812.07865
    """
    tau = np.zeros(len(t))
    mask = np.where((t >= win1) & (t <= win2))

    frac = (t[mask] - win1) / (win2 - win1)

    tau[mask] = np.sin(0.5 * np.pi * frac) ** 2
    mask = np.where(t > win2)
    tau[mask] = 1
    return tau


def build_hybrid(wf_pn, wf_nr, dt, dphi, dpsi, win1, win2, delta_t=None):
    """
    delta_t: output delta_t for hybrid, default None. If None then will use PN delta_t/4
    """
    wf_pn = copy.deepcopy(wf_pn)
    wf_nr = copy.deepcopy(wf_nr)

    (wf_nr.apply_time_shift(-dt).apply_phase_shift(dphi).apply_polarisation_shift(dpsi))

    # loop over all modes in waveform dict
    # apply hybridisation to each mode
    # todo: apply global time and phase shift

    modes = wf_nr.modes

    # use 2,2 mode to get common time array
    wf_pn_x = wf_pn.times
    wf_nr_x_shifted = wf_nr.times

    t_start = np.min([wf_pn_x[0], wf_nr_x_shifted[0]])
    t_end = np.max([wf_pn_x[-1], wf_nr_x_shifted[-1]])
    if delta_t is None:
        delta_t = (wf_pn_x[1] - wf_pn_x[0]) / 4
    times = np.arange(t_start, t_end, delta_t)

    # window/blending function
    tau = blending_function(times, win1, win2)

    wf_hybrid = {}
    wf_hybrid["hlm"] = {}

    for i, mode in enumerate(modes):
        ell = mode[0]
        mm = mode[1]

        # get PN waveform
        wf_pn_x = wf_pn.times
        wf_pn_y = wf_pn.hlms[ell, mm]

        # get NR waveform and apply time, phase and polarisation shifts
        wf_nr_x_shifted = wf_nr.times
        wf_nr_y_shifted = wf_nr.hlms[ell, mm]

        # interpolate real and imag PN and NR waveforms
        # by setting the extrapolation behaviour like this
        # we can easily zero pad the left and right to the same time array
        # and then easily multiply the window function

        hybrid_pn_real = IUS(wf_pn_x, wf_pn_y.real, ext=1)(times) * (1 - tau)
        hybrid_nr_real = IUS(wf_nr_x_shifted, wf_nr_y_shifted.real, ext=1)(times) * tau
        hybrid_real = hybrid_pn_real + hybrid_nr_real

        hybrid_pn_imag = IUS(wf_pn_x, wf_pn_y.imag, ext=1)(times) * (1 - tau)
        hybrid_nr_imag = IUS(wf_nr_x_shifted, wf_nr_y_shifted.imag, ext=1)(times) * tau
        hybrid_imag = hybrid_pn_imag + hybrid_nr_imag

        wf_hybrid["hlm"][mode] = hybrid_real + 1.0j * hybrid_imag

    # time shift peak sum
    # amps = []
    # for mode in modes:
    #     amps.append(np.abs(wf_hybrid['hlm'][mode]))
    # amps = np.array(amps)
    # t_peak = times[np.argmax(amps.sum(0))]

    # wf_hybrid['t'] = times - t_peak
    wf_hybrid["t"] = times

    wf_hybrid = waveform.Waveform(wf_hybrid["t"], hlms=wf_hybrid["hlm"])

    return wf_hybrid
